"""Plugin for generic approval flows."""

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import AppRegistryNotReady
from django.core.validators import MinValueValidator
from django.utils.translation import gettext_lazy as _

from order.models import PurchaseOrder
from order.views import PurchaseOrderDetail
from users.models import Owner

from InvenTree.helpers_model import notify_responsible
from InvenTree.status_codes import PurchaseOrderStatus
from plugin import InvenTreePlugin
from plugin.events import trigger_event
from plugin.mixins import (
    AppMixin,
    NavigationMixin,
    PanelMixin,
    SettingsContentMixin,
    SettingsMixin,
    UrlsMixin,
)

from .notifications import ApprovalReminder, ApprovalStarted


def update_reminder_task(*args, **kwargs):
    """Ensures reminder task is up to date."""
    try:
        from django_q.models import Schedule
    except AppRegistryNotReady:  # pragma: no cover
        return
    from InvenTree.tasks import schedule_task
    from plugin import registry as plugin_registry

    plg = plugin_registry.get_plugin(ApprovalPlugin.SLUG)
    if not plg:
        return

    val = plg.get_setting("REMINDER")
    if val and int(val) > 0:
        # Schedule
        schedule_task(
            "inventree_approval.ApprovalPlugin.send_reminder",
            schedule_type=Schedule.MINUTES,
            minutes=val,
            repeats=1,
        )
    else:
        # Remove
        Schedule.objects.filter(
            func="inventree_approval.ApprovalPlugin.send_reminder"
        ).delete()


class ApprovalPlugin(
    PanelMixin,
    AppMixin,
    SettingsMixin,
    UrlsMixin,
    NavigationMixin,
    SettingsContentMixin,
    InvenTreePlugin,
):
    """Plugin for generic approval flows."""

    NAME = "ApprovalPlugin"
    SLUG = "inventree_approval"
    TITLE = "Approvals"
    SETTINGS = {
        "APPROVAL_MINIMUM_APPROVERS": {
            "name": _("Minimum Approvers"),
            "description": _("Minimum number of approvers required for approval"),
            "default": 2,
            "validator": [int, MinValueValidator(1)],
        },
        "APPROVAL_MAXIMUM_DENIERS": {
            "name": _("Maximum Deniers"),
            "description": _("Maximum number of deniers required for denial"),
            "default": 1,
            "validator": [int, MinValueValidator(1)],
        },
        "REMINDER": {
            "name": _("Reminder interval"),
            "description": _(
                "How often should reminders be evaluated (in minutes; 0 == never)"
            ),
            "default": 0,
            "validator": [int, MinValueValidator(0)],
            "after_save": update_reminder_task,
        },
    }
    MIN_VERSION = "0.13.0"
    NAVIGATION_TAB_NAME = "Approvals"
    NAVIGATION_TAB_ICON = "far fa-check-square"
    NAVIGATION = [
        {"name": "Overview", "link": f"plugin:{SLUG}:approval-ui"},
    ]

    def __init__(self):
        """Init a plugin."""
        super().__init__()

        update_reminder_task()

    def get_settings_content(self, request):
        """Custom settings content for the plugin."""
        from .templatetags.inventree_approval import approval_jsx

        tag_data = approval_jsx()
        return f"""
        <h3>Settings</h3>
        {tag_data}
        <div id="root"></div>
        """

    def setup_urls(self):
        """Setup url endpoints for this plugin."""
        from .api import urlpatterns

        return urlpatterns

    def get_custom_panels(self, view, request):
        """Custom panels for the plugin."""
        panels = []

        # This panel will *only* display on the PurchaseOrder view,
        if isinstance(view, PurchaseOrderDetail):
            panels.append({
                "title": "Approval",
                "icon": "fa-check-square",
                "content_template": "inventree_approval/panel.html",
            })
        return panels


# define transition method
try:
    from generic.states import TransitionMethod
except ImportError:  # pragma: no cover

    class TransitionMethod:
        """Dummy class for transition method."""


contenttype_obj = ContentType.objects.get_for_model(PurchaseOrder)


class POchanges(TransitionMethod):
    """Transition method for PurchaseOrder objects."""

    def transition(current_state, target_state, instance, default_action, **kwargs):  # noqa: N805
        """Example override function for state transition."""
        from .models import Approval

        if not isinstance(instance, PurchaseOrder):
            return False

        # Not placed already
        if instance.can_cancel:
            # is_active = getSetting("PURCHASEORDER_REQUIRE_APPROVAL")
            is_active = True

            if (
                is_active
                and Approval.objects.filter(
                    object_id=instance.id, content_type=contenttype_obj
                ).count()
                == 0
            ):
                if instance.responsible is None:
                    instance.responsible = Owner.create(obj=instance.created_by)

                approval = Approval.objects.create(
                    name=f"PO: {instance.reference}",
                    content_object=instance,
                    created_by=instance.created_by,
                    modified_by=instance.created_by,
                    responsible=instance.responsible,
                )

                instance.status = 15  # PurchaseOrderStatus.PENDING_APPROVAL.value
                instance.save()

                trigger_event("purchaseorder.approval_start", id=instance.pk)

                # Send email to responsible to link to approval page
                try:
                    notify_responsible(
                        approval,
                        Approval,
                        content=ApprovalStarted,
                    )
                except TypeError:
                    pass

                return instance
            else:
                # Run the default action
                return False

        # Is being placed right now
        if current_state == 15 and target_state == 20:
            # Run the default action

            # TODO remove this
            instance.status = PurchaseOrderStatus.PENDING.value
            instance.save()

            return False


def send_reminder():
    """Send reminders for open approvals."""
    from .models import Approval, ApprovalStateGroups

    # Get all open approvals
    to_send = Approval.objects.filter(
        status__in=ApprovalStateGroups.OPEN, responsible__isnull=False
    )

    # Send reminder to responsible
    for approval in to_send:
        notify_responsible(
            approval,
            Approval,
            content=ApprovalReminder,
        )

    # TODO: Send reminder to requested approvers
