"""API views for the Approval app."""

from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.urls import include, path, re_path
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.generic import TemplateView

from rest_framework import permissions, serializers, status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from users.serializers import OwnerSerializer

from InvenTree.api import MetadataView
from InvenTree.filters import SEARCH_ORDER_FILTER_ALIAS
from InvenTree.helpers import str2bool
from InvenTree.helpers_model import notify_users
from InvenTree.mixins import (
    CreateAPI,
    ListAPI,
    ListCreateAPI,
    RetrieveUpdateAPI,
    RetrieveUpdateDestroyAPI,
)
from InvenTree.serializers import InvenTreeModelSerializer, UserSerializer

from .helpers import CustomDjangoModelPermissions, TaggedObjectRelatedField
from .models import Approval, ApprovalDecision
from .notifications import ApprovalDecisionRequested


class PermMixin:
    """Mixin to add custom permission classes to the API views."""

    permission_classes = [
        CustomDjangoModelPermissions,
    ]


class UserDetailApiMixin:
    """Mixin to add extra context information regarding user/owner details to endpoint serializers."""

    # from users.authentication import ApiTokenAuthentication
    # authentication_classes = [ApiTokenAuthentication]

    def get_serializer(self, *args, **kwargs):
        """Add extra context information to the endpoint serializer."""
        try:
            user_detail = str2bool(self.request.GET.get("user_detail", None))
        except AttributeError:
            user_detail = None

        kwargs["user_detail"] = user_detail
        return super().get_serializer(*args, **kwargs)

    def get_object(self):
        """Get the object to be viewed."""
        # Use default behaviour if no type is specified
        if self.kwargs.get("type", None) is None:
            return super().get_object()

        # Alternatively, use the generic relationship to get the object
        queryset = self.filter_queryset(self.get_queryset())
        obj = get_object_or_404(
            queryset,
            **{
                "content_type__model": self.kwargs["type"],
                "object_id": self.kwargs["pk"],
            },
        )
        self.check_object_permissions(self.request, obj)
        return obj


class UserDetailSerializerMixin:
    """Mixin to determine if extra serializer fields are required."""

    def __init__(self, *args, **kwargs):
        """Determine if extra serializer fields are required."""
        user_detail = kwargs.pop("user_detail", False)

        super().__init__(*args, **kwargs)

        if user_detail is not True:
            for field in self.Meta.user_detail_fields:
                self.fields.pop(field)


class ApprovalDecisionSerializer(UserDetailSerializerMixin, InvenTreeModelSerializer):
    """Serializes an ApprovalDecision object."""

    user_detail = UserSerializer(source="user", read_only=True, many=False)

    class Meta:
        """Meta data for ApprovalDecisionSerializer."""

        model = ApprovalDecision
        exclude = [
            "metadata",
        ]
        user_detail_fields = [
            "user_detail",
        ]

    def is_valid(self, *, raise_exception=False):
        """Insert user to save request."""
        request = self.context["request"]
        self.initial_data["user"] = request.user.pk
        return super().is_valid(raise_exception=raise_exception)


class ApprovalSerializer(UserDetailSerializerMixin, InvenTreeModelSerializer):
    """Serializes an Approval object."""

    status_text = serializers.CharField(source="get_status_display", read_only=True)
    content_object = TaggedObjectRelatedField(read_only=True)
    content_model = serializers.SerializerMethodField(
        method_name="get_content_model", read_only=True
    )
    model = serializers.CharField(required=False, write_only=True)
    decisions = ApprovalDecisionSerializer(many=True, read_only=True, user_detail=True)
    creation_date = serializers.DateTimeField(format="iso-8601", required=False)
    created_by_detail = UserSerializer(source="created_by", read_only=True, many=False)
    modified_date = serializers.DateTimeField(format="iso-8601", required=False)
    modified_by_detail = UserSerializer(
        source="modified_by", read_only=True, many=False
    )
    finalised_date = serializers.DateTimeField(format="iso-8601", required=False)
    finalised_by_detail = UserSerializer(
        source="finalised_by", read_only=True, many=False
    )
    responsible_detail = OwnerSerializer(source="responsible", read_only=True)
    owner_detail = OwnerSerializer(source="owner", read_only=True)

    def get_content_model(self, obj):
        """Get the content model name."""
        return obj.content_type.model

    class Meta:
        """Meta data for ApprovalSerializer."""

        model = Approval
        exclude = [
            "reference_int",
            "metadata",
            "data",
            # 'object_id',
            # 'content_type',
        ]

        read_only_fields = [
            "creation_date",
            "finalised_date",
            "modified_date",
        ]

        user_detail_fields = [
            "created_by_detail",
            "modified_by_detail",
            "finalised_by_detail",
            "responsible_detail",
            "owner_detail",
        ]

    def is_valid(self, *, raise_exception=False):
        """Overwritten validation method to patch content_type and name field."""
        # If model is provided, use it to determine content_type
        if "model" in self.initial_data:
            model = self.initial_data.pop("model")
            mdl_splt = model.split(".")
            if len(mdl_splt) != 2:
                raise serializers.ValidationError({
                    "model": _("Invalid model name provided.")
                })
            content_type = ContentType.objects.get(
                app_label=mdl_splt[0], model=mdl_splt[1]
            )
            self.initial_data["content_type"] = content_type.pk

        # If name is not provided, use the object name
        if "name" not in self.initial_data or self.initial_data["name"] == "":
            obj = ContentType.objects.get(
                pk=self.initial_data["content_type"]
            ).get_object_for_this_type(id=self.initial_data["object_id"])
            self.initial_data["name"] = str(_(f"Approval for {str(obj)}"))
        return super().is_valid(raise_exception=raise_exception)


class ApprovalList(PermMixin, UserDetailApiMixin, ListCreateAPI):
    """API endpoint for listing all Approval objects."""

    filter_backends = SEARCH_ORDER_FILTER_ALIAS
    queryset = Approval.objects.all()
    serializer_class = ApprovalSerializer
    filterset_fields = [
        "owner",
        "status",
        "content_type",
        "object_id",
    ]
    search_fields = [
        "name",
        "description",
        "reference",
    ]
    ordering_field_aliases = {
        "reference": ["reference_int", "reference"],
    }
    ordering_fields = [
        "name",
        "status",
        "finalised_date",
        "modified_date",
        "creation_date",
        "reference",
        "finalised",
    ]
    ordering = "-reference"

    def clean_data(self, data: dict) -> dict:
        """Custom clean_data method adds current user."""
        data = super().clean_data(data)
        data["created_by"] = self.request.user.pk
        return data

    def options(self, request, *args, **kwargs):  # pragma: no cover
        """Return metadata for the API endpoint."""
        try:
            return super().options(request, *args, **kwargs)
        except AttributeError:
            pass
        return Response({}, status=status.HTTP_200_OK)


class ApprovalDetail(PermMixin, UserDetailApiMixin, RetrieveUpdateDestroyAPI):
    """API endpoint for detail view of an Approval object."""

    queryset = Approval.objects.all()
    serializer_class = ApprovalSerializer


class ApprovalDecisionList(PermMixin, UserDetailApiMixin, ListCreateAPI):
    """API endpoint for listing all ApprovalDecision objects."""

    queryset = ApprovalDecision.objects.all()
    serializer_class = ApprovalDecisionSerializer
    filterset_fields = [
        "approval",
        "user",
        "status",
    ]
    search_fields = [
        "comment",
    ]
    ordering_fields = [
        "approval",
        "user",
        "status",
        "date",
    ]

    def list(self, request, *args, **kwargs):
        """List the ApprovalDecision objects."""
        self.kwargs = kwargs
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        """Get the queryset for the ApprovalDecision list."""
        queryset = super().get_queryset()
        return queryset.filter(approval=self.kwargs["pk"])


class ApprovalDecisionDetail(PermMixin, UserDetailApiMixin, RetrieveUpdateDestroyAPI):
    """API endpoint for detail view of an ApprovalDecision object."""

    queryset = ApprovalDecision.objects.all()
    serializer_class = ApprovalDecisionSerializer


class ApprovalApproveSerializer(InvenTreeModelSerializer):
    """Serializes an ApprovalDecision object."""

    class Meta:
        """Meta data for ApprovalDecisionSerializer."""

        model = ApprovalDecision
        exclude = [
            "metadata",
        ]

    def is_valid(self, *, raise_exception=False):
        """Insert data to save request."""
        request = self.context["request"]
        self.initial_data["user"] = request.user.pk
        self.initial_data["approval"] = request.parser_context["kwargs"].get("pk", None)
        self.initial_data["decision"] = True
        return super().is_valid(raise_exception=raise_exception)


class ApproveView(PermMixin, CreateAPI):
    """API endpoint to approve approval."""

    queryset = ApprovalDecision.objects.all()
    serializer_class = ApprovalApproveSerializer


class RequestApprovalSerializer(serializers.Serializer):
    """Serializes an ApprovalDecision object."""

    requested_user_pk = serializers.IntegerField(required=True)
    requested_user = UserSerializer(read_only=True, many=False)

    def is_valid(self, *, raise_exception=False):
        """Insert data to save request."""
        request = self.context["request"]
        self.initial_data["requested_by"] = request.user.pk
        self.initial_data["requested_date"] = now()
        self.initial_data["approval"] = request.parser_context["kwargs"].get("pk", None)
        ret = super().is_valid(raise_exception=raise_exception)
        return ret


class RequestApproveView(PermMixin, CreateAPI):
    """API endpoint to request approval."""

    queryset = ApprovalDecision.objects.all()
    serializer_class = RequestApprovalSerializer

    def perform_create(self, serializer):
        """Create a new approval request."""
        data = serializer.initial_data
        approval = get_object_or_404(Approval, pk=data["approval"])
        usr_rq = get_object_or_404(User, pk=data["requested_user_pk"])
        ret = approval.request_decision(user=self.request.user, requested_user=usr_rq)
        notify_users(
            [
                usr_rq,
            ],
            approval,
            Approval,
            content=ApprovalDecisionRequested,
        )
        return ret


PERM_REF = [
    "view_approval",
    "add_approval",
    "change_approval",
    "delete_approval",
    "view_approvaldecision",
    "add_approvaldecision",
    "change_approvaldecision",
    "delete_approvaldecision",
]


def get_perms():
    """Get the current permissions for the approval app."""
    return {
        a.codename: [b.id for b in a.group_set.all()]
        for a in Permission.objects.filter(codename__in=PERM_REF)
    }


class PermissionSerializer(serializers.Serializer):
    """Serializes an Permission object."""

    id = serializers.IntegerField(read_only=True)  # noqa: A003
    name = serializers.CharField(read_only=True)
    view_approval = serializers.BooleanField(default=False)
    add_approval = serializers.BooleanField(default=False)
    change_approval = serializers.BooleanField(default=False)
    delete_approval = serializers.BooleanField(default=False)
    view_approvaldecision = serializers.BooleanField(default=False)
    add_approvaldecision = serializers.BooleanField(default=False)
    change_approvaldecision = serializers.BooleanField(default=False)
    delete_approvaldecision = serializers.BooleanField(default=False)

    def update(self, instance, validated_data):
        """Update the permission set for a group."""
        diffs = [a for a in PERM_REF if instance[0][a] is not validated_data[a]]
        for diff in diffs:
            permission_obj = Permission.objects.get(codename=diff)
            group_obj = Group.objects.get(pk=instance[0]["id"])

            if validated_data[diff]:
                group_obj.permissions.add(permission_obj)
            else:
                group_obj.permissions.remove(permission_obj)

            group_obj.save()

        if diffs:
            instance[0].update(validated_data)
            return instance[0]
        return instance


class PermissionList(PermMixin, ListAPI, RetrieveUpdateAPI):
    """API endpoint for listing permission objects."""

    serializer_class = PermissionSerializer
    permission_classes = [permissions.IsAdminUser]

    def list(self, request, *args, **kwargs):  # noqa: A003
        """List the permissions."""
        self.kwargs = kwargs
        ret = super().list(request, *args, **kwargs)
        if "pk" in kwargs:
            ret.data = ret.data[0]
        return ret

    def get_queryset(self):
        """Get the queryset for the permission list."""
        kwargs = getattr(self, "kwargs", {})
        all_groups = (
            [
                get_object_or_404(Group, pk=kwargs["pk"]),
            ]
            if "pk" in kwargs
            else Group.objects.all()
        )
        groups = get_perms()
        data = [
            {
                "id": a.id,
                "name": a.name,
                **{x: a.id in groups[x] for x in PERM_REF},
            }
            for a in all_groups
        ]
        return data

    def get_object(self):
        """Get the object for the permission list - needed for metadata."""
        return self.get_queryset()


detail_api = [
    path(
        "decision/",
        include([
            path(
                "<int:pk>/",
                ApprovalDecisionDetail.as_view(),
                name="api-approval-decision-detail",
            ),
            path(
                "",
                ApprovalDecisionList.as_view(),
                name="api-approval-decision-list",
            ),
        ]),
    ),
    path("approve/", ApproveView.as_view(), name="api-approval-approve"),
    path("request/", RequestApproveView.as_view(), name="api-approval-request"),
    path(
        "metadata/",
        MetadataView.as_view(),
        {"model": Approval},
        name="api-approval-metadata",
    ),
    path("", ApprovalDetail.as_view(), name="api-approval-detail"),
]


approval_api_urls = [
    path(
        "permissions/",
        include([
            path("<int:pk>/", PermissionList.as_view(), name="api-permission-detail"),
            path("", PermissionList.as_view(), name="api-permission-list"),
        ]),
    ),
    path("<str:type>:<int:pk>/", include(detail_api)),
    path("<int:pk>/", include(detail_api)),
    path("", ApprovalList.as_view(), name="api-approval-list"),
]

urlpatterns = [
    path("api/", include(approval_api_urls)),
    re_path(
        ".*",
        ensure_csrf_cookie(
            TemplateView.as_view(template_name="inventree_approval/index.html")
        ),
        name="approval-ui",
    ),
]
